@@ -77,3 +77,54 @@ int uint64_compare(uint64 x, uint64 y)
 	return x.lo < y.lo ? -1 : +1;
     return 0;
 }
+
+uint64 uint64_subtract(uint64 x, uint64 y)
+{
+    x.lo -= y.lo;
+    x.hi -= y.hi + (x.lo > ~y.lo ? 1 : 0);
+    return x;
+}
+
+double uint64_to_double(uint64 x)
+{
+    return (4294967296.0 * x.hi) + (double)x.lo;
+}
+
+uint64 uint64_shift_right(uint64 x, int shift)
+{
+    if (shift < 32) {
+	x.lo >>= shift;
+	x.lo |= (x.hi << (32-shift));
+	x.hi >>= shift;
+    } else {
+	x.lo = x.hi >> (shift-32);
+	x.hi = 0;
+    }
+    return x;
+}
+
+uint64 uint64_shift_left(uint64 x, int shift)
+{
+    if (shift < 32) {
+	x.hi <<= shift;
+	x.hi |= (x.lo >> (32-shift));
+	x.lo <<= shift;
+    } else {
+	x.hi = x.lo << (shift-32);
+	x.lo = 0;
+    }
+    return x;
+}
+
+uint64 uint64_from_decimal(char *str)
+{
+    uint64 ret;
+    ret.hi = ret.lo = 0;
+    while (*str >= '0' && *str <= '9') {
+	ret = uint64_add(uint64_shift_left(ret, 3),
+			 uint64_shift_left(ret, 1));
+	ret = uint64_add32(ret, *str - '0');
+	str++;
+    }
+    return ret;
+}
