@@ -43,7 +43,13 @@ static void logwrite(struct LogContext *ctx, void *data, int len)
 	bufchain_add(&ctx->queue, data, len);
     } else if (ctx->state == L_OPEN) {
 	assert(ctx->lgfp);
-	fwrite(data, 1, len, ctx->lgfp);
+	if (fwrite(data, 1, len, ctx->lgfp) < len) {
+	    logfclose(ctx);
+	    ctx->state = L_ERROR;
+	    /* Log state is L_ERROR so this won't cause a loop */
+	    logevent(ctx->frontend,
+		     "Disabled writing session log due to error while writing");
+	}
     }				       /* else L_ERROR, so ignore the write */
 }
 
@@ -85,7 +91,7 @@ static void logfopen_callback(void *handle, int mode)
 	ctx->state = L_ERROR;	       /* disable logging */
     } else {
 	fmode = (mode == 1 ? "ab" : "wb");
-	ctx->lgfp = f_open(ctx->currlogfilename, fmode);
+	ctx->lgfp = f_open(ctx->currlogfilename, fmode, FALSE);
 	if (ctx->lgfp)
 	    ctx->state = L_OPEN;
 	else
@@ -101,8 +107,9 @@ static void logfopen_callback(void *handle, int mode)
     }
 
     event = dupprintf("%s session log (%s mode) to file: %s",
-		      (mode == 0 ? "Disabled writing" :
-                       mode == 1 ? "Appending" : "Writing new"),
+		      ctx->state == L_ERROR ?
+		      (mode == 0 ? "Disabled writing" : "Error writing") :
+		      (mode == 1 ? "Appending" : "Writing new"),
 		      (ctx->cfg.logtype == LGTYP_ASCII ? "ASCII" :
 		       ctx->cfg.logtype == LGTYP_DEBUG ? "raw" :
 		       ctx->cfg.logtype == LGTYP_PACKETS ? "SSH packets" :
@@ -149,7 +156,7 @@ void logfopen(void *handle)
     /* substitute special codes in file name */
     xlatlognam(&ctx->currlogfilename, ctx->cfg.logfilename,ctx->cfg.host, &tm);
 
-    ctx->lgfp = f_open(ctx->currlogfilename, "r");  /* file already present? */
+    ctx->lgfp = f_open(ctx->currlogfilename, "r", FALSE);  /* file already present? */
     if (ctx->lgfp) {
 	fclose(ctx->lgfp);
 	if (ctx->cfg.logxfovr != LGXF_ASK) {
@@ -204,6 +211,9 @@ void log_eventlog(void *handle, const char *event)
 	fprintf(stderr, "%s\n", event);
 	fflush(stderr);
     }
+    /* If we don't have a context yet (eg winnet.c init) then skip entirely */
+    if (!ctx)
+	return;
     if (ctx->cfg.logtype != LGTYP_PACKETS &&
 	ctx->cfg.logtype != LGTYP_SSHRAW)
 	return;
@@ -217,8 +227,9 @@ void log_eventlog(void *handle, const char *event)
  * Set of blanking areas must be in increasing order.
  */
 void log_packet(void *handle, int direction, int type,
-		char *texttype, void *data, int len,
-		int n_blanks, const struct logblank_t *blanks)
+		char *texttype, const void *data, int len,
+		int n_blanks, const struct logblank_t *blanks,
+		const unsigned long *seq)
 {
     struct LogContext *ctx = (struct LogContext *)handle;
     char dumpdata[80], smalldata[5];
@@ -230,13 +241,20 @@ void log_packet(void *handle, int direction, int type,
 	return;
 
     /* Packet header. */
-    if (texttype)
-        logprintf(ctx, "%s packet type %d / 0x%02x (%s)\r\n",
-                  direction == PKT_INCOMING ? "Incoming" : "Outgoing",
-                  type, type, texttype);
-    else
+    if (texttype) {
+	if (seq) {
+	    logprintf(ctx, "%s packet #0x%lx, type %d / 0x%02x (%s)\r\n",
+		      direction == PKT_INCOMING ? "Incoming" : "Outgoing",
+		      *seq, type, type, texttype);
+	} else {
+	    logprintf(ctx, "%s packet type %d / 0x%02x (%s)\r\n",
+		      direction == PKT_INCOMING ? "Incoming" : "Outgoing",
+		      type, type, texttype);
+	}
+    } else {
         logprintf(ctx, "%s raw data\r\n",
                   direction == PKT_INCOMING ? "Incoming" : "Outgoing");
+    }
 
     /*
      * Output a hex/ASCII dump of the packet body, blanking/omitting
@@ -352,7 +370,7 @@ void log_reconfig(void *handle, Config *cfg)
  * translate format codes into time/date strings
  * and insert them into log file name
  *
- * "&Y":YYYY   "&m":MM   "&d":DD   "&T":hhmm   "&h":<hostname>   "&&":&
+ * "&Y":YYYY   "&m":MM   "&d":DD   "&T":hhmmss   "&h":<hostname>   "&&":&
  */
 static void xlatlognam(Filename *dest, Filename src,
 		       char *hostname, struct tm *tm) {
@@ -373,7 +391,7 @@ static void xlatlognam(Filename *dest, Filename src,
 	    char c;
 	    s++;
 	    size = 0;
-	    if (*s) switch (c = *s++, tolower(c)) {
+	    if (*s) switch (c = *s++, tolower((unsigned char)c)) {
 	      case 'y':
 		size = strftime(buf, sizeof(buf), "%Y", tm);
 		break;
