@@ -1,4 +1,4 @@
-/* $Id: mac_res.r 6513 2006-01-08 18:18:11Z jacob $ */
+/* $Id: mac_res.r 8848 2010-01-17 17:27:27Z jacob $ */
 /*
  * Copyright (c) 1999, 2002, 2003 Ben Harris
  * All rights reserved.
@@ -1221,7 +1221,7 @@ resource 'DITL' (wAbout, "about", purgeable) {
 	StaticText { disabled, "PuTTY"},
 	{ 42, 13, 74, 227 },
 	StaticText { disabled, "Some version or other\n"
-			       "Copyright  1997-2006 Simon Tatham"},
+			       "Copyright  1997-2010 Simon Tatham"},
     }
 };
 
@@ -1242,12 +1242,12 @@ type 'TEXT' {
 };
 
 resource 'TEXT' (wLicence, "licence", purgeable) {
-    "PuTTY is copyright 1997-2006 Simon Tatham.\n"
+    "PuTTY is copyright 1997-2010 Simon Tatham.\n"
     "\n"
     "Portions copyright Robert de Bath, Joris van Rantwijk, Delian "
     "Delchev, Andreas Schultz, Jeroen Massar, Wez Furlong, Nicolas Barry, "
     "Justin Bradford, Ben Harris, Malcolm Smith, Ahmad Khalifa, Markus "
-    "Kuhn, and CORE SDI S.A.\n"
+    "Kuhn, Colin Watson, and CORE SDI S.A.\n"
     "\n"    
     "Permission is hereby granted, free of charge, to any person "
     "obtaining a copy of this software and associated documentation "
