@@ -1,4 +1,4 @@
-/* $Id: macctrls.c 5434 2005-03-02 23:33:59Z owen $ */
+/* $Id: macctrls.c 6815 2006-08-28 10:35:12Z simon $ */
 /*
  * Copyright (c) 2003 Ben Harris
  * All rights reserved.
@@ -2327,6 +2327,17 @@ int dlg_coloursel_results(union control *ctrl, void *dlg,
 	return 0;
 }
 
+void dlg_label_change(union control *ctrl, void *dlg, char const *text)
+{
+    /*
+     * This function is currently only used by the config box to
+     * switch the labels on the host and port boxes between serial
+     * and network modes. Since the Mac port does not have a serial
+     * back end, this function can safely do nothing.
+     */
+}
+
+
 /*
  * Local Variables:
  * c-file-style: "simon"
