@@ -1,4 +1,4 @@
-/* $Id: macpgen.r 6513 2006-01-08 18:18:11Z jacob $ */
+/* $Id: macpgen.r 8392 2009-01-05 23:49:19Z jacob $ */
 /*
  * Copyright (c) 1999, 2002, 2003 Ben Harris
  * All rights reserved.
@@ -422,7 +422,7 @@ resource 'DITL' (wAbout, "about", purgeable) {
 	StaticText { disabled, "PuTTYgen"},
 	{ 42, 13, 74, 227 },
 	StaticText { disabled, "Some version or other\n"
-			       "Copyright  1997-2006 Simon Tatham"},
+			       "Copyright  1997-2009 Simon Tatham"},
     }
 };
 
@@ -443,12 +443,12 @@ type 'TEXT' {
 };
 
 resource 'TEXT' (wLicence, "licence", purgeable) {
-    "Copyright 1997-2006 Simon Tatham.\n"
+    "Copyright 1997-2009 Simon Tatham.\n"
     "\n"
     "Portions copyright Robert de Bath, Joris van Rantwijk, Delian "
     "Delchev, Andreas Schultz, Jeroen Massar, Wez Furlong, Nicolas Barry, "
     "Justin Bradford, Ben Harris, Malcolm Smith, Ahmad Khalifa, Markus "
-    "Kuhn, and CORE SDI S.A.\n"
+    "Kuhn, Colin Watson, and CORE SDI S.A.\n"
     "\n"    
     "Permission is hereby granted, free of charge, to any person "
     "obtaining a copy of this software and associated documentation "
