@@ -794,6 +794,11 @@
     return term_data(term, is_stderr, data, len);
 }
 
+- (int)fromBackendUntrusted:(const char *)data len:(int)len
+{
+    return term_data_untrusted(term, data, len);
+}
+
 - (void)startAlert:(NSAlert *)alert
     withCallback:(void (*)(void *, int))callback andCtx:(void *)ctx
 {
@@ -885,6 +890,11 @@
     // FIXME: else show restart menu item
 }
 
+- (Terminal *)term
+{
+    return term;
+}
+
 @end
 
 int from_backend(void *frontend, int is_stderr, const char *data, int len)
@@ -893,6 +903,12 @@ int from_backend(void *frontend, int is_stderr, const char *data, int len)
     return [win fromBackend:data len:len isStderr:is_stderr];
 }
 
+int from_backend_untrusted(void *frontend, const char *data, int len)
+{
+    SessionWindow *win = (SessionWindow *)frontend;
+    return [win fromBackendUntrusted:data len:len];
+}
+
 int get_userpass_input(prompts_t *p, unsigned char *in, int inlen)
 {
     SessionWindow *win = (SessionWindow *)p->frontend;
@@ -922,7 +938,7 @@ void ldisc_update(void *frontend, int echo, int edit)
 
 char *get_ttymode(void *frontend, const char *mode)
 {
-    SessionWindow *win = (SessionWindow *)ctx;
+    SessionWindow *win = (SessionWindow *)frontend;
     Terminal *term = [win term];
     return term_get_ttymode(term, mode);
 }
