@@ -4,7 +4,8 @@
 #
 # Pass an argument of the form `2004-02-08' to have the archive
 # tagged as a development snapshot; of the form `0.54' to have it
-# tagged as a release.
+# tagged as a release; of the form `r1234' to have it tagged as a
+# custom build. Otherwise it'll be tagged as unidentified.
 
 case "$1" in
   ????-??-??)
@@ -13,6 +14,11 @@ case "$1" in
     ver="-DSNAPSHOT=$1"
     docver=
     ;;
+  r*)
+    arcsuffix="-$1"
+    ver="-DSVN_REV=$1"
+    docver=
+    ;;
   '')
     arcsuffix=
     ver=
