@@ -9,15 +9,23 @@
 #include <assert.h>
 #include <errno.h>
 #include <ctype.h>
+#include <limits.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <pwd.h>
 #include "putty.h"
 #include "storage.h"
 #include "tree234.h"
 
+#ifdef PATH_MAX
+#define FNLEN PATH_MAX
+#else
+#define FNLEN 1024 /* XXX */
+#endif
+
 enum {
     INDEX_DIR, INDEX_HOSTKEYS, INDEX_HOSTKEYS_TMP, INDEX_RANDSEED,
     INDEX_SESSIONDIR, INDEX_SESSION,
@@ -80,36 +88,78 @@ static char *unmungestr(const char *in)
     return ret;
 }
 
-static void make_filename(char *filename, int index, const char *subname)
+static char *make_filename(int index, const char *subname)
 {
-    char *home;
-    int len;
-    home = getenv("HOME");
-    if (!home)
-        home="/";
-    strncpy(filename, home, FILENAME_MAX);
-    len = strlen(filename);
+    char *env, *tmp, *ret;
+
+    /*
+     * Allow override of the PuTTY configuration location, and of
+     * specific subparts of it, by means of environment variables.
+     */
+    if (index == INDEX_DIR) {
+	struct passwd *pwd;
+
+	env = getenv("PUTTYDIR");
+	if (env)
+	    return dupstr(env);
+	env = getenv("HOME");
+	if (env)
+	    return dupprintf("%s/.putty", env);
+	pwd = getpwuid(getuid());
+	if (pwd && pwd->pw_dir)
+	    return dupprintf("%s/.putty", pwd->pw_dir);
+	return dupstr("/.putty");
+    }
+    if (index == INDEX_SESSIONDIR) {
+	env = getenv("PUTTYSESSIONS");
+	if (env)
+	    return dupstr(env);
+	tmp = make_filename(INDEX_DIR, NULL);
+	ret = dupprintf("%s/sessions", tmp);
+	sfree(tmp);
+	return ret;
+    }
     if (index == INDEX_SESSION) {
         char *munged = mungestr(subname);
-        char *fn = dupprintf("/.putty/sessions/%s", munged);
-        strncpy(filename + len, fn, FILENAME_MAX - len);
-        sfree(fn);
-        sfree(munged);
-    } else {
-        strncpy(filename + len,
-                index == INDEX_DIR ? "/.putty" :
-                index == INDEX_SESSIONDIR ? "/.putty/sessions" :
-                index == INDEX_HOSTKEYS ? "/.putty/sshhostkeys" :
-                index == INDEX_HOSTKEYS_TMP ? "/.putty/sshhostkeys.tmp" :
-                index == INDEX_RANDSEED ? "/.putty/randomseed" :
-                "/.putty/ERROR", FILENAME_MAX - len);
+	tmp = make_filename(INDEX_SESSIONDIR, NULL);
+	ret = dupprintf("%s/%s", tmp, munged);
+	sfree(tmp);
+	sfree(munged);
+	return ret;
     }
-    filename[FILENAME_MAX-1] = '\0';
+    if (index == INDEX_HOSTKEYS) {
+	env = getenv("PUTTYSSHHOSTKEYS");
+	if (env)
+	    return dupstr(env);
+	tmp = make_filename(INDEX_DIR, NULL);
+	ret = dupprintf("%s/sshhostkeys", tmp);
+	sfree(tmp);
+	return ret;
+    }
+    if (index == INDEX_HOSTKEYS_TMP) {
+	tmp = make_filename(INDEX_HOSTKEYS, NULL);
+	ret = dupprintf("%s.tmp", tmp);
+	sfree(tmp);
+	return ret;
+    }
+    if (index == INDEX_RANDSEED) {
+	env = getenv("PUTTYRANDOMSEED");
+	if (env)
+	    return dupstr(env);
+	tmp = make_filename(INDEX_DIR, NULL);
+	ret = dupprintf("%s/randomseed", tmp);
+	sfree(tmp);
+	return ret;
+    }
+    tmp = make_filename(INDEX_DIR, NULL);
+    ret = dupprintf("%s/ERROR", tmp);
+    sfree(tmp);
+    return ret;
 }
 
 void *open_settings_w(const char *sessionname, char **errmsg)
 {
-    char filename[FILENAME_MAX];
+    char *filename;
     FILE *fp;
 
     *errmsg = NULL;
@@ -121,18 +171,24 @@ void *open_settings_w(const char *sessionname, char **errmsg)
      * other error will trip us up later on so there's no real need
      * to catch it now.
      */
-    make_filename(filename, INDEX_DIR, sessionname);
-    mkdir(filename, 0700);
-    make_filename(filename, INDEX_SESSIONDIR, sessionname);
-    mkdir(filename, 0700);
+    filename = make_filename(INDEX_SESSIONDIR, NULL);
+    if (mkdir(filename, 0700) != 0) {
+	char *filename2 = make_filename(INDEX_DIR, NULL);
+	mkdir(filename2, 0700);
+	sfree(filename2);
+	mkdir(filename, 0700);
+    }
+    sfree(filename);
 
-    make_filename(filename, INDEX_SESSION, sessionname);
+    filename = make_filename(INDEX_SESSION, sessionname);
     fp = fopen(filename, "w");
     if (!fp) {
         *errmsg = dupprintf("Unable to create %s: %s",
                             filename, strerror(errno));
+	sfree(filename);
 	return NULL;                   /* can't open */
     }
+    sfree(filename);
     return fp;
 }
 
@@ -226,13 +282,14 @@ const char *get_setting(const char *key)
 
 void *open_settings_r(const char *sessionname)
 {
-    char filename[FILENAME_MAX];
+    char *filename;
     FILE *fp;
     char *line;
     tree234 *ret;
 
-    make_filename(filename, INDEX_SESSION, sessionname);
+    filename = make_filename(INDEX_SESSION, sessionname);
     fp = fopen(filename, "r");
+    sfree(filename);
     if (!fp)
 	return NULL;		       /* can't open */
 
@@ -305,7 +362,35 @@ int read_setting_i(void *handle, const char *key, int defvalue)
 
 int read_setting_fontspec(void *handle, const char *name, FontSpec *result)
 {
-    return !!read_setting_s(handle, name, result->name, sizeof(result->name));
+    /*
+     * In GTK1-only PuTTY, we used to store font names simply as a
+     * valid X font description string (logical or alias), under a
+     * bare key such as "Font".
+     * 
+     * In GTK2 PuTTY, we have a prefix system where "client:"
+     * indicates a Pango font and "server:" an X one; existing
+     * configuration needs to be reinterpreted as having the
+     * "server:" prefix, so we change the storage key from the
+     * provided name string (e.g. "Font") to a suffixed one
+     * ("FontName").
+     */
+    char *suffname = dupcat(name, "Name", NULL);
+    if (read_setting_s(handle, suffname, result->name, sizeof(result->name))) {
+	sfree(suffname);
+	return TRUE;		       /* got new-style name */
+    }
+    sfree(suffname);
+
+    /* Fall back to old-style name. */
+    memcpy(result->name, "server:", 7);
+    if (!read_setting_s(handle, name,
+			result->name + 7, sizeof(result->name) - 7) ||
+	!result->name[7]) {
+	result->name[0] = '\0';
+	return FALSE;
+    } else {
+	return TRUE;
+    }
 }
 int read_setting_filename(void *handle, const char *name, Filename *result)
 {
@@ -314,7 +399,14 @@ int read_setting_filename(void *handle, const char *name, Filename *result)
 
 void write_setting_fontspec(void *handle, const char *name, FontSpec result)
 {
-    write_setting_s(handle, name, result.name);
+    /*
+     * read_setting_fontspec had to handle two cases, but when
+     * writing our settings back out we simply always generate the
+     * new-style name.
+     */
+    char *suffname = dupcat(name, "Name", NULL);
+    write_setting_s(handle, suffname, result.name);
+    sfree(suffname);
 }
 void write_setting_filename(void *handle, const char *name, Filename result)
 {
@@ -341,18 +433,20 @@ void close_settings_r(void *handle)
 
 void del_settings(const char *sessionname)
 {
-    char filename[FILENAME_MAX];
-    make_filename(filename, INDEX_SESSION, sessionname);
+    char *filename;
+    filename = make_filename(INDEX_SESSION, sessionname);
     unlink(filename);
+    sfree(filename);
 }
 
 void *enum_settings_start(void)
 {
     DIR *dp;
-    char filename[FILENAME_MAX];
+    char *filename;
 
-    make_filename(filename, INDEX_SESSIONDIR, NULL);
+    filename = make_filename(INDEX_SESSIONDIR, NULL);
     dp = opendir(filename);
+    sfree(filename);
 
     return dp;
 }
@@ -362,19 +456,22 @@ char *enum_settings_next(void *handle, char *buffer, int buflen)
     DIR *dp = (DIR *)handle;
     struct dirent *de;
     struct stat st;
-    char fullpath[FILENAME_MAX];
-    int len;
+    char *fullpath;
+    int maxlen, thislen, len;
     char *unmunged;
 
-    make_filename(fullpath, INDEX_SESSIONDIR, NULL);
-    len = strlen(fullpath);
+    fullpath = make_filename(INDEX_SESSIONDIR, NULL);
+    maxlen = len = strlen(fullpath);
 
     while ( (de = readdir(dp)) != NULL ) {
-        if (len < FILENAME_MAX) {
-            fullpath[len] = '/';
-            strncpy(fullpath+len+1, de->d_name, FILENAME_MAX-(len+1));
-            fullpath[FILENAME_MAX-1] = '\0';
-        }
+        thislen = len + 1 + strlen(de->d_name);
+	if (maxlen < thislen) {
+	    maxlen = thislen;
+	    fullpath = sresize(fullpath, maxlen+1, char);
+	}
+	fullpath[len] = '/';
+	strncpy(fullpath+len+1, de->d_name, thislen - (len+1));
+	fullpath[thislen] = '\0';
 
         if (stat(fullpath, &st) < 0 || !S_ISREG(st.st_mode))
             continue;                  /* try another one */
@@ -383,9 +480,11 @@ char *enum_settings_next(void *handle, char *buffer, int buflen)
         strncpy(buffer, unmunged, buflen);
         buffer[buflen-1] = '\0';
         sfree(unmunged);
+	sfree(fullpath);
         return buffer;
     }
 
+    sfree(fullpath);
     return NULL;
 }
 
@@ -408,12 +507,13 @@ int verify_host_key(const char *hostname, int port,
 		    const char *keytype, const char *key)
 {
     FILE *fp;
-    char filename[FILENAME_MAX];
+    char *filename;
     char *line;
     int ret;
 
-    make_filename(filename, INDEX_HOSTKEYS, NULL);
+    filename = make_filename(INDEX_HOSTKEYS, NULL);
     fp = fopen(filename, "r");
+    sfree(filename);
     if (!fp)
 	return 1;		       /* key does not exist */
 
@@ -478,7 +578,7 @@ void store_host_key(const char *hostname, int port,
     FILE *rfp, *wfp;
     char *newtext, *line;
     int headerlen;
-    char filename[FILENAME_MAX], tmpfilename[FILENAME_MAX];
+    char *filename, *tmpfilename;
 
     newtext = dupprintf("%s@%d:%s %s\n", keytype, port, hostname, key);
     headerlen = 1 + strcspn(newtext, " ");   /* count the space too */
@@ -486,18 +586,22 @@ void store_host_key(const char *hostname, int port,
     /*
      * Open both the old file and a new file.
      */
-    make_filename(tmpfilename, INDEX_HOSTKEYS_TMP, NULL);
+    tmpfilename = make_filename(INDEX_HOSTKEYS_TMP, NULL);
     wfp = fopen(tmpfilename, "w");
     if (!wfp) {
-        char dir[FILENAME_MAX];
+        char *dir;
 
-        make_filename(dir, INDEX_DIR, NULL);
+        dir = make_filename(INDEX_DIR, NULL);
         mkdir(dir, 0700);
+	sfree(dir);
+
         wfp = fopen(tmpfilename, "w");
     }
-    if (!wfp)
+    if (!wfp) {
+	sfree(tmpfilename);
 	return;
-    make_filename(filename, INDEX_HOSTKEYS, NULL);
+    }
+    filename = make_filename(INDEX_HOSTKEYS, NULL);
     rfp = fopen(filename, "r");
 
     /*
@@ -521,17 +625,20 @@ void store_host_key(const char *hostname, int port,
 
     rename(tmpfilename, filename);
 
+    sfree(tmpfilename);
+    sfree(filename);
     sfree(newtext);
 }
 
 void read_random_seed(noise_consumer_t consumer)
 {
     int fd;
-    char fname[FILENAME_MAX];
+    char *fname;
 
-    make_filename(fname, INDEX_RANDSEED, NULL);
+    fname = make_filename(INDEX_RANDSEED, NULL);
     fd = open(fname, O_RDONLY);
-    if (fd) {
+    sfree(fname);
+    if (fd >= 0) {
 	char buf[512];
 	int ret;
 	while ( (ret = read(fd, buf, sizeof(buf))) > 0)
@@ -543,9 +650,9 @@ void read_random_seed(noise_consumer_t consumer)
 void write_random_seed(void *data, int len)
 {
     int fd;
-    char fname[FILENAME_MAX];
+    char *fname;
 
-    make_filename(fname, INDEX_RANDSEED, NULL);
+    fname = make_filename(INDEX_RANDSEED, NULL);
     /*
      * Don't truncate the random seed file if it already exists; if
      * something goes wrong half way through writing it, it would
@@ -553,10 +660,12 @@ void write_random_seed(void *data, int len)
      */
     fd = open(fname, O_CREAT | O_WRONLY, 0600);
     if (fd < 0) {
-	char dir[FILENAME_MAX];
+	char *dir;
 
-	make_filename(dir, INDEX_DIR, NULL);
+	dir = make_filename(INDEX_DIR, NULL);
 	mkdir(dir, 0700);
+	sfree(dir);
+
 	fd = open(fname, O_CREAT | O_WRONLY, 0600);
     }
 
@@ -568,6 +677,7 @@ void write_random_seed(void *data, int len)
     }
 
     close(fd);
+    sfree(fname);
 }
 
 void cleanup_all(void)
